/* tclsave.c */

/* Vis5D version 5.0 */

/*
Vis5D system for visualizing five dimensional gridded data sets
Copyright (C) 1990 - 1997 Bill Hibbard, Johan Kellum, Brian Paul,
Dave Santek, and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * Tcl-based save and restore.
 */



#include <stdio.h>
#include <string.h>
#include "globals.h"
#include "graphics.h"
#include "gui.h"





/*
 * Return the parameters which describe a colorbar curve.
 * If one or more color table entries were 'drawn' by the user, rather
 * than computed from the parameters, return 1.  Return 0 if the user
 * has not drawn the curves.
 * Input:  index - context index
 *         graphic - one of:  VIS5D_ISOSURF, VIS5D_CHSLICE, VIS5D_CVSLICE,
 *                   VIS5D_VOLUME, VIS5D_TRAJ or VIS5D_TOPO.
 *         var - variable number if type isn't VIS5D_TOPO
 * Output: params - pointer to address of parameter array
 */
static int get_colorbar_params( int index, int graphic, int vindex, int var,
                                float params[] )
{
   float *p;
   int i;
   unsigned int temptable[256], *table;
   int same;

   vis5d_get_color_table_params( index, graphic, vindex, var, &p ); 
   for (i=0; i<7; i++) params[i] = p[i];

   /*
    * This is tricky.  Compare the graphic's color table to one computed
    * from its parameters.  If they're different return 1, else return 0.
    * This is used by the SAVE function to determine if all the color
    * table entries have to be written or just the parameters which
    * describe the curves.
    */
   vis5d_get_color_table_address( index, graphic, vindex, var, &table ); 
   vis5d_color_table_recompute( temptable, 256, params, 1, 1 );
   same = 1;
   for (i=0;i<255;i++) {
      if (table[i]!=temptable[i]) {
         same = 0;
         break;
      }
   }

   if (same) {
      return 0;
   }
   else {
      return 1;
   }
}





/*
 * Save current graphics and colors to the 'savefile' as a Tcl script.
 * Input:  savefile - filename to save to.
 * Return:  0 for success,
 *          VIS5D_BAD_VALUE if unable to open file
 *          VIS5D_FAIL if error while writing file.
 */
int tcl_save( int index, char *savefile )
{
   FILE *f;
   int cyo, chowmany, cwhichones[VIS5D_MAX_CONTEXTS];
   int time, var, i, k, set;
   int numvars;
   float r, g, b, a;
   char varname[20];

   vis5d_get_num_of_ctxs_in_display( index, &chowmany, cwhichones);

   f = fopen(savefile,"w");
   if (!f) {
      return VIS5D_BAD_VALUE;
   }  

   /* Prolog */ 
   fprintf(f,"#Vis5D 4.3 Tcl save file\n\n");


   for ( cyo = 0; cyo < chowmany; cyo++){
      int vindex = cwhichones[cyo];

      vis5d_get_numvars( cwhichones[cyo], &numvars );
      /* misc colors */
      fprintf(f,"\n#Box color\n");
      vis5d_get_color( index, VIS5D_BOX, 0, &r, &g, &b, &a );
      fprintf(f,"vis5d_set_color %d VIS5D_BOX 0 %5.3f %5.3f %5.3f %5.3f\n",
              index, r,g,b,a );

      fprintf(f,"\n#Light map color\n");
      vis5d_get_color( index, VIS5D_LIGHT_MAP, 0, &r, &g, &b, &a );
      fprintf(f,"vis5d_set_color %d VIS5D_LIGHT_MAP 0 %5.3f %5.3f %5.3f %5.3f\n",
              index, r,g,b,a );

      fprintf(f,"\n#Dark map color\n");
      vis5d_get_color( index, VIS5D_DARK_MAP, 0, &r, &g, &b, &a );
      fprintf(f,"vis5d_set_color %d VIS5D_DARK_MAP 0 %5.3f %5.3f %5.3f %5.3f\n",
              index, r,g,b,a );

      fprintf(f,"\n#Background color\n");
      vis5d_get_color( index, VIS5D_BACKGROUND, 0, &r, &g, &b, &a );
      fprintf(f,
              "vis5d_set_color  %d VIS5D_BACKGROUND 0 %5.3f %5.3f %5.3f %5.3f\n",
              index, r,g,b,a );

      /* Text labels */
      fprintf(f,"\n#Text labels\n");
      {
         int i = 1;
         int x, y;
         char label[1000];
         while (vis5d_get_label( index, i, &x, &y, label )==0) {
            fprintf(f,"vis5d_make_label %d %d %d \"%s\"\n", index, x, y, label );
            i++;
         }
      }

      
      /* View matrix */
      fprintf(f,"\n#Viewing matrix\n");
      {
         float mat[4][4];
         int i, j;
         vis5d_get_matrix( index, mat );
         fprintf(f,"vis5d_set_matrix %d {", index);
         for (i=0;i<4;i++) {
            for (j=0;j<4;j++) {
               fprintf(f," %g", mat[i][j] );
            }
         }
         fprintf(f," }\n");
      }

      /* Camera */
      fprintf(f,"\n#Camera\n");
      {
         int perspec;
         float front, zoom;
         vis5d_get_camera( index, &perspec, &front, &zoom );
         fprintf(f,"vis5d_set_camera %d %d %g %g\n", index, perspec, front, zoom );
      }

      /* Cloned and computed physical variables */
      fprintf(f,"\n#Cloned or computed variables\n");
      for (var=0; var<numvars; var++) {
         int type;
         vis5d_get_var_type( cwhichones[cyo], var, &type );
         if (type==VIS5D_CLONE) {
            int vartoclone;
            char origname[20], clonename[20];
            vis5d_get_var_name( cwhichones[cyo], var, origname );
            vis5d_get_var_info( cwhichones[cyo], var, (void*) &vartoclone );
            vis5d_get_var_name( cwhichones[cyo], vartoclone, clonename );
            fprintf(f,"vis5d_make_clone_variable %d \"%s\" \"%s\"\n",
                    cwhichones[cyo], origname, clonename );
         }
         else if (type==VIS5D_EXT_FUNC) {
            char funcname[100];
            vis5d_get_var_info( cwhichones[cyo], var, (void*) funcname );
            fprintf(f,"vis5d_compute_ext_func %d \"%s\"\n", index, funcname );
         }
         else if (type==VIS5D_EXPRESSION) {
            char expr[100];
            vis5d_get_var_info( cwhichones[cyo], var, (void*) expr );
            fprintf(f,"vis5d_make_expr_var %d \"%s\"\n", index, expr );
         }
      }

      /* Isosurfaces */
      fprintf(f,"\n#Isosurfaces\n");
      for (var=0; var<numvars; var++) {
         float isolevel, min, max;
         int colorvarowner, colorvar;
         vis5d_get_isosurface( cwhichones[cyo], var, &isolevel );
         vis5d_get_var_range( cwhichones[cyo], var, &min, &max );
         vis5d_get_isosurface_color_var( cwhichones[cyo], var, &colorvarowner, &colorvar ); 
         vis5d_get_var_name( cwhichones[cyo], var, varname );
         if (isolevel!=min) {
            fprintf(f,"vis5d_set_isosurface %d \"%s\" %g\n", cwhichones[cyo], varname, isolevel);
            if (colorvar>-1) {
               char colorvarname[20];
               vis5d_get_var_name( colorvarowner, colorvar, colorvarname );
               fprintf(f,"vis5d_set_isosurface_color_var %d \"%s\" %d  \"%s\"\n",
                       cwhichones[cyo], varname, colorvarowner, colorvarname );
            }
            /* command to recompute the isosurface */
            fprintf(f,"vis5d_make_isosurface %d  VIS5D_ALL_TIMES \"%s\" 0\n",
                    cwhichones[cyo], varname );
         }
         if (vis5d_enable_graphics(cwhichones[cyo],VIS5D_ISOSURF,var,VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_ISOSURF %d VIS5D_ON\n",
                       cwhichones[cyo], var);
         }
      }

      /* Horizontal contour slices */
      fprintf(f,"\n#Horizontal contour slices\n");
      for (var=0;var<numvars;var++) {
         float interval, low, high, level;
         vis5d_get_hslice( vindex, var, &interval, &low, &high, &level );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,"vis5d_set_hslice %d \"%s\" %g %g %g %g\n",vindex, varname,
                 interval, low, high, level );
         if (vis5d_enable_graphics(vindex, VIS5D_HSLICE, var, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_HSLICE %d VIS5D_ON\n",
                       vindex, var);
         }
      }

      /* Vertical contour slices */
      fprintf(f,"\n#Vertical contour slices\n");
      for (var=0;var<numvars;var++) {
         float interval, low, high, r0, c0, r1, c1;
         vis5d_get_vslice( vindex, var, &interval, &low, &high, &r0,&c0, &r1,&c1 );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,"vis5d_set_vslice %d \"%s\" %g %g %g  %g %g  %g %g\n", vindex,
                 varname, interval, low, high, r0,c0, r1,c1 );
         if (vis5d_enable_graphics(vindex, VIS5D_VSLICE, var, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_VSLICE %d VIS5D_ON\n",
                       vindex, var);
         }      
      }

      /* Horizontal colored slices */
      fprintf(f,"\n#Horizontal colored slices\n");
      for (var=0;var<numvars;var++) {
         float level;
         vis5d_get_chslice( vindex, var, &level );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,"vis5d_set_chslice %d \"%s\" %g\n",vindex, varname, level );
         if (vis5d_enable_graphics(vindex, VIS5D_CHSLICE, var, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_CHSLICE %d VIS5D_ON\n",
                       vindex, var);
         }      
      }
      
      /* Vertical colored slices */
      fprintf(f,"\n#Vertical colored slices\n");
      for (var=0;var<numvars;var++) {
         float level, r0, c0, r1, c1;
         vis5d_get_cvslice( vindex, var, &r0, &c0, &r1, &c1 );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,"vis5d_set_cvslice %d \"%s\" %g %g %g %g\n",vindex,
                 varname, r0, c0, r1, c1 );
         if (vis5d_enable_graphics(vindex, VIS5D_CVSLICE, var, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_CVSLICE %d VIS5D_ON\n",
                       vindex, var);
         }      
      }

      /* Horizontal wind vector slices */
      fprintf(f,"\n#Horizontal wind vector slices\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         float density, scale, level;
         vis5d_get_hwindslice( index, i, &density, &scale, &level );
         fprintf(f,"vis5d_set_hwindslice %d %d  %g %g %g\n",index,  i, density,
                 scale, level );
         if (vis5d_enable_graphics(index, VIS5D_HWIND, i, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_HWIND %d VIS5D_ON\n",
                       index, i);
         }
      }

      /* Vertical wind vector slices */
      fprintf(f,"\n#Vertical wind vector slices\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         float density, scale, r0, c0, r1, c1;
         vis5d_get_vwindslice( index, i, &density, &scale, &r0, &c0, &r1, &c1 );
         fprintf(f,"vis5d_set_vwindslice %d %d  %g %g  %g %g %g %g\n",index,  i,
                 density, scale, r0, c0, r1, c1 );
         if (vis5d_enable_graphics(index, VIS5D_VWIND, i, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_VWIND %d VIS5D_ON\n",
                       index, i);
         }
      }

      /* Horizontal wind stream slices */
      fprintf(f,"\n#Horizontal stream slices\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         float density, level;
         vis5d_get_hstreamslice( index, i, &density, &level );
         fprintf(f,"vis5d_set_hstreamslice %d %d  %g %g\n",index,  i, density,
                 level );
         if (vis5d_enable_graphics(index, VIS5D_HSTREAM, i, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_HSTREAM %d VIS5D_ON\n",
                       index, i);      
         }            
      }

      /* Vertical wind stream slices */
      fprintf(f,"\n#Vertical stream slices\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         float density, r0, c0, r1, c1;
         vis5d_get_vstreamslice( index, i, &density, &r0, &c0, &r1, &c1 );
         fprintf(f,"vis5d_set_vstreamslice %d %d  %g %g %g %g %g\n",index,  i,
                 density, r0, c0, r1, c1 );
         if (vis5d_enable_graphics(index, VIS5D_VSTREAM, i, VIS5D_GET)){
            fprintf(f,"vis5d_enable_graphics %d VIS5D_VSTREAM %d VIS5D_ON\n",
                       index, i);
         }      
      }

      /* Trajectories */
      fprintf(f,"\n#Trajectories\n");
      {
         float prevstep = 0.0, prevlength = 0.0;
         int prevribbon = 0;
         int numtraj = vis5d_get_num_traj( index );
         for (i=0;i<numtraj;i++) {
            float row, col, lev, step, length;
            int timestep, group, ribbon;
            vis5d_get_traj_info( index, i, &row, &col, &lev, &timestep,
                                 &step, &length, &group, &ribbon );
            if (i==0 || step!=prevstep || length!=prevlength
                || ribbon!=prevribbon) {
               fprintf(f,"vis5d_set_traj %d %g %g %d\n", index, step, length, ribbon );
            }
            prevstep = step;
            prevlength = length;
            prevribbon = ribbon;
            
            fprintf(f,"vis5d_make_traj %d %g %g %g %d %d\n", index, row, col, lev,
                    timestep, group );
         }
         for (i=0; i<VIS5D_TRAJ_SETS; i++){
            if (vis5d_enable_graphics(index, VIS5D_TRAJ, i, VIS5D_GET)){
               fprintf(f,"vis5d_enable_graphics %d VIS5D_TRAJ %d VIS5D_ON\n",
                          index, i);      
            }
         }            
      }


      /* Isosurface colors */
      fprintf(f,"\n#Isosurface colors\n");
      for (var=0;var<numvars;var++) {
         vis5d_get_color( index, VIS5D_ISOSURF, vindex*MAXVARS+var, &r, &g, &b, &a );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,
           "vis5d_set_color %d VIS5D_ISOSURF %d \"%s\" %5.3f %5.3f %5.3f %5.3f\n", index,
           vindex, varname, r,g,b,a );
      }

      /* HSlice colors */
      fprintf(f,"\n#Horizontal contour slice colors\n");
      for (var=0;var<numvars;var++) {
         vis5d_get_color( index, VIS5D_HSLICE, vindex*MAXVARS+var, &r, &g, &b, &a );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,
             "vis5d_set_color %d VIS5D_HSLICE %d \"%s\" %5.3f %5.3f %5.3f %5.3f\n",
             index, vindex, varname, r,g,b,a );
      }

      /* VSlice colors */
      fprintf(f,"\n#Vertical contour slice colors\n");
      for (var=0;var<numvars;var++) {
         vis5d_get_color( index, VIS5D_VSLICE, vindex*MAXVARS+var, &r, &g, &b, &a );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,
            "vis5d_set_color %d VIS5D_VSLICE %d \"%s\" %5.3f %5.3f %5.3f %5.3f\n",
            index, vindex, varname, r,g,b,a );
      }

      /* Colored HSlice colors */
      fprintf(f,"\n#Horizontal colored slice tickmark colors\n");
      for (var=0;var<numvars;var++) {
         vis5d_get_color( index, VIS5D_CHSLICE, vindex*MAXVARS+var, &r, &g, &b, &a );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,
           "vis5d_set_color %d VIS5D_CHSLICE %d \"%s\" %5.3f %5.3f %5.3f %5.3f\n",
           index, vindex, varname, r,g,b,a );
      }

      /* Colored VSlice colors */
      fprintf(f,"\n#Vertical colored slice tickmark colors\n");
      for (var=0;var<numvars;var++) {
         vis5d_get_color( index, VIS5D_CVSLICE, vindex*MAXVARS+var, &r, &g, &b, &a );
         vis5d_get_var_name( vindex, var, varname );
         fprintf(f,
          "vis5d_set_color %d VIS5D_CVSLICE %d \"%s\" %5.3f %5.3f %5.3f %5.3f\n",
          index, vindex, varname, r,g,b,a );
      }

      /* HWind colors */
      fprintf(f,"\n#Horizontal wind slice colors\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         vis5d_get_color( index, VIS5D_HWIND, i, &r, &g, &b, &a );
         fprintf(f,
                 "vis5d_set_color %d VIS5D_HWIND %d %5.3f %5.3f %5.3f %5.3f\n",
                 index, i, r,g,b,a );
      }

      /* VWind colors */
      fprintf(f,"\n#Vertical wind slice colors\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         vis5d_get_color( index, VIS5D_VWIND, i, &r, &g, &b, &a );
         fprintf(f,
                 "vis5d_set_color %d VIS5D_VWIND %d %5.3f %5.3f %5.3f %5.3f\n",
                 index, i, r,g,b,a );
      }

      /* Horizontal Stream colors */
      fprintf(f,"\n#Horizontal stream slice colors\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         vis5d_get_color( index, VIS5D_HSTREAM, i, &r, &g, &b, &a );
         fprintf(f,
                 "vis5d_set_color %d VIS5D_HSTREAM %d %5.3f %5.3f %5.3f %5.3f\n",
                 index, i, r,g,b,a );
      }

      /* Vertical Stream colors */
      fprintf(f,"\n#Vertical stream slice colors\n");
      for (i=0;i<VIS5D_WIND_SLICES;i++) {
         vis5d_get_color( index, VIS5D_VSTREAM, i, &r, &g, &b, &a );
         fprintf(f,
                 "vis5d_set_color %d VIS5D_VSTREAM %d %5.3f %5.3f %5.3f %5.3f\n",
                 index, i, r,g,b,a );
      }

      /* Trajectory colors */
      fprintf(f,"\n#Trajectory colors\n");
      for (i=0;i<VIS5D_TRAJ_SETS;i++) {
         int colorvarowner, colorvar;
         vis5d_get_color( index, VIS5D_TRAJ, i, &r, &g, &b, &a );
         fprintf(f,"vis5d_set_color %d VIS5D_TRAJ %d %5.3f %5.3f %5.3f %5.3f\n",
                 index, i, r,g,b,a );
         vis5d_get_trajectory_color_var( index, i, &colorvarowner, &colorvar ); 
         if (colorvar>=0) {
            char varname[20];
            vis5d_get_var_name( colorvarowner, colorvar, varname );
            fprintf(f,"vis5d_set_trajectory_color_var %d %d %d \"%s\"\n",
                    index,i, colorvarowner, varname );
         }
      }  



      /* Isosurface color tables */
      fprintf(f,"\n#Isosurface color tables\n");
      for (var=0;var<numvars;var++) {
         unsigned int *ctable;
         char varname[20];
         float params[8];

         vis5d_get_var_name( index, var, varname );

         k = get_colorbar_params( index, VIS5D_ISOSURF, vindex, var, params );

         fprintf(f,"vis5d_set_color_table_params %d VIS5D_ISOSURF %d", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0], params[1],
                 params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_ISOSURF, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_ISOSURF %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }         


      /* Horizontal color slice color tables */
      fprintf(f,"\n#Horizontal color slice color tables\n");
      for (var=0;var<numvars;var++) {
         unsigned int *ctable;
         char varname[20];
         float params[8];

         vis5d_get_var_name( index, var, varname );

         k = get_colorbar_params( index, VIS5D_CHSLICE, vindex, var, params );

         fprintf(f,"vis5d_set_color_table_params %d VIS5D_CHSLICE %d ", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0], params[1],
                 params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_CHSLICE, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_CHSLICE %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }         

      /* Vertical color slice color tables */
      fprintf(f,"\n#Vertical color slice color tables\n");
      for (var=0;var<numvars;var++) {
         unsigned int *ctable;
         char varname[20];
         float params[8];

         vis5d_get_var_name( index, var, varname );

         k = get_colorbar_params( index, VIS5D_CVSLICE, vindex, var, params );

         fprintf(f,"vis5d_set_color_table_params %d VIS5D_CVSLICE %d", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0], params[1],
                 params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_CVSLICE, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_CVSLICE %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }         

      /* Volume color tables */
      fprintf(f,"\n#Volume color tables\n");
      for (var=0;var<numvars;var++) {
         unsigned int *ctable;
         char varname[20];
         float params[8];

         vis5d_get_var_name( index, var, varname );

         k = get_colorbar_params( index, VIS5D_VOLUME, vindex, var, params );

         fprintf(f,"vis5d_set_color_table_params %d VIS5D_VOLUME %d", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0], params[1],
                 params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_VOLUME, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_VOLUME %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }         

      /* Trajectory color tables */
      fprintf(f,"\n#Trajectory color tables\n");
      for (var=0;var<numvars;var++) {
         unsigned int *ctable;
         float params[8];

         vis5d_get_var_name( index, var, varname );
         k = get_colorbar_params( index, VIS5D_TRAJ, vindex, var, params );

         fprintf(f,"vis5d_set_color_table_params %d VIS5D_TRAJ %d", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0], params[1],
                 params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_TRAJ, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_TRAJ %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }         


      /* Topography color tables */
      fprintf(f,"\n#Topography color tables\n");
      for (var=-1;var<numvars;var++) {
         unsigned int *ctable;
         float params[8];

         if (var>=0) {
            vis5d_get_var_name( index, var, varname );
            k = get_colorbar_params( index, VIS5D_TOPO, vindex, var, params );
         }
         else {
            int j;
            for (j=0; j<8; j++) params[j] = 0;
            strcpy( varname, "-1" );
            k = 1;
         }
         fprintf(f,"vis5d_set_color_table_params %d VIS5D_TOPO %d", index, vindex);
         fprintf(f," \"%s\" %.3f %.3f %.3f %.3f\n", varname, params[0],
                     params[1], params[2], params[3] );

         if (k) {
            /* the color table can't be described by the parameters alone */
            /* save each individual table entry */
            vis5d_get_color_table_address( index, VIS5D_TOPO, vindex, var, &ctable ); 
            for (i=0;i<256;i++) {
               fprintf(f,"vis5d_set_color_table_entry %d VIS5D_TOPO %d", index, vindex);
               fprintf(f," \"%s\" %d %d %d %d %d\n", varname, i, 
                       UNPACK_RED(ctable[i]), UNPACK_GREEN(ctable[i]),
                       UNPACK_BLUE(ctable[i]), UNPACK_ALPHA(ctable[i]) );
            }
            fprintf(f,"\n");
         }
      }
   }         
   {
      int colorvar, colorvarowner;
      vis5d_get_topo_color_var( index, &colorvarowner, &colorvar ); 
      fprintf(f, "\n");
      fprintf(f, "vis5d_set_topo_color_var %d %d %d\n", index, colorvarowner, colorvar );
   }

   /* misc vis5d_graphics_mode stuff */
   {
      if (vis5d_graphics_mode(index, VIS5D_BOX, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_BOX VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_BOX VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_CLOCK, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_CLOCK VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_CLOCK VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_MAP, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_MAP VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_MAP VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_TOPO, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_TOPO VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_TOPO VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_TEXTURE, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_TEXTURE VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_TEXTURE VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_JULIAN, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_JULIAN VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_JULIAN VIS5D_OFF\n", index);
      }
   
      if (vis5d_graphics_mode(index, VIS5D_BARBS, VIS5D_GET)){
         fprintf(f,"vis5d_graphics_mode %d VIS5D_BARBS VIS5D_ON\n", index);
      }
      else{
         fprintf(f,"vis5d_graphics_mode %d VIS5D_BARBS VIS5D_OFF\n", index);
      }
   
      /* there could be more of these vis5d_graphics_mode enable calls */
      /* but I'll wait till some want's them in there */
   } 




   fprintf(f, "\nvis5d_draw_frame %d\n", index );

   fclose(f);
   return 0;
}


